/* 
    This program reads a map database file in the CIA World Base format and
   produces a netmap-format output file.
 */


#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <strings.h>
#include <sys/file.h>
#include "cbdmap.h"

#define	BIT32	int		/* change these definitions to suit your */
#define	BIT16	unsigned short	/* own machine				 */
#define	BIT8	char
#define	TRUE	1
#define	FALSE	0
/* 
    The data-compression scheme uses integer seconds to represent lat/long,
   and stores each stroke as a [dx,dy] from the previous point. If dy will
   fit into 8 bits and dx into 7 bits, then the entire [dx,dy] is stored in a
   16-bit field with bit 0x4000 turned on as a flag. If either value is too
   large for that scheme, then both are stored as 32-bit values, with the
   0x40000000 bit turned off (even in negative numbers) in the first of them.
 */

#define	MAXSEG	20000		/* maximum strokes in a segment */
#define	MAXSEGS 10000

#define abs(x)	(x<0 ? -x : x)
#define sign(x) ((x>=0) ? 1 : -1)

struct seghead sb;

BIT16	segbuf[MAXSEG];

struct segdict sd[MAXSEGS];

int	MapDetail=99;

/* ------------------------------------------------------------------------ */

main(argc, argv)
int	argc;
char	**argv;
{
	extern int optind;
	extern char *optarg;
	extern char *strcpy();
	int	c, initdone,badopts;

	initdone = 0;
	badopts = 0;
	while ((c = getopt(argc, argv, "d:m")) != EOF)
	    switch (c) {

/* -d: detail. Argument is an integer giving map detail to include. */
		case 'd': sscanf(optarg,"%d",&MapDetail);
			  break;
/* -m: Ascii map files follow. "-m f2 f2 f3"			 */
		case 'm': for(;optind<argc && argv[optind][0]!='-';optind++) {
			     if (!initdone) ; 
			     initdone++;
			     TranslateMap(argv[optind]);
			  }
			  break;
		case '?': badopts++;
	    }
	    fflush(stdout);
	    if (badopts) {
                exit (2);
            }
}


/* ------------------------------------------------------------------------ */

TranslateMap(name)
    char *name;
{
	FILE	*df;
	char	line[100],
		of[100],
		lad,		/* 'N' or 'S' */
		lod;		/* 'E' or 'W' */
	int	Ofile,
		first,
		lsn,irnk,nop,
		ladg,lam,las,lodg,lom,los,
		ipn,		/* polygon number */
		big,small,
		lnum,
		nc,i,j,
		strokecount,
		segmax,segsize,
		segcount,bytecount,
		segsmall;
	BIT32	jlat,		/* integer latitude in seconds */
		jlong,		/* integer longitude in seconds */
		haveLat=0,
		haveLong=0,
		dx,dy;
	struct cbdhead header;

	if (!(df = fopen(name, "r"))) {
	    perror(name);
	    return;
	}
	strcpy(of,name);
	if (index(of,'.')) *(index(of,'.')) = '\0';
	strcat(of,".cbd");

	Ofile = open(of, O_RDWR|O_CREAT|O_TRUNC, 0666);
	if (Ofile < 0) {
	    perror(of);
	    return;
	}

	fprintf(stderr,"Processing map file %s\n",name);

	header.magic = 0;		/* It's not magic until we are done */
	
	write(Ofile, &header, sizeof (header));	/* write out dummy header */
	big = 0; small = 0;

	segcount = 0;
	segmax = 0;

/* The header line tells us  the serial number, rank, and path size */
	lnum = 0;
	while (fgets(line, 60, df)) {
	    lnum++;
	    for(nc=0; nc<18&&line[nc]!=NULL; nc++);

	    if (nc<18) continue;

	    sscanf(line, "%7d%2d%6d", &lsn, &irnk, &nop);

	    if (irnk>MapDetail) {
	        for (i = 1; i <= nop; i++) {fgets(line, 60, df); lnum++;}
	    } else {
		sb.id = lsn;
		first = FALSE;
		segcount++;
		if (segcount > MAXSEGS) {
		    fprintf(stderr,"Segment count overflow.\n");
		    close(Ofile); fclose(df);
		    return;
		}
		sd[segcount].absaddr = tell(Ofile);
		sd[segcount].segid = lsn;
		sd[segcount].rank = irnk;

		sd[segcount].minlat = 99999999;
		sd[segcount].maxlat = -99999999;
		sd[segcount].minlong = 999999999;
		sd[segcount].maxlong = -99999999;

		bytecount = 0;
		segsize = 0;
	        for (i = 1; i <= nop; i++) {
		    fgets(line, 60, df);
		    lnum++;
		    for(j = 0; j <= 19; j++) {
			if (line[j]==' ') line[j]='0';
		    }
#define DIG(x) ((int)(x)-48)
		    ladg=10*DIG(line[0])+DIG(line[1]);
		    lam=10*DIG(line[2])+DIG(line[3]);
		    las=10*DIG(line[4])+DIG(line[5]);
		    lad=line[6];
		    lodg=100*DIG(line[7])+10*DIG(line[8])+DIG(line[9]);
		    lom=10*DIG(line[10])+DIG(line[11]);
		    los=10*DIG(line[12])+DIG(line[13]);
		    lod=line[14];
		    ipn=1000*DIG(line[16])+100*DIG(line[17])+10*DIG(line[18])+DIG(line[19]);

		    if (ipn != i) {
			fprintf(stderr,"Data error at sequence %d, segment %d: line %d expected but %d found:\n%s\n",lnum, lsn, i, ipn, line);
			if (ipn > 0 && ipn < nop) i=ipn;
		    }


		    jlat = 3600*ladg + 60*lam + las;
 		    if (lad == 'S')
			    jlat = -jlat;
		    jlong = 3600*lodg + 60*lom + los;
		    if (lod == 'W')
			    jlong = -jlong;

		    if (sd[segcount].maxlat < jlat) sd[segcount].maxlat = jlat;
		    if (sd[segcount].minlat > jlat) sd[segcount].minlat = jlat;
		    if (sd[segcount].maxlong < jlong) sd[segcount].maxlong = jlong;
		    if (sd[segcount].minlong > jlong) sd[segcount].minlong = jlong;
		    if (!first) {
			first = TRUE;
			sb.orgx = jlong;
			sb.orgy = jlat;
			strokecount = 0;
		    } else {
		    	dx = jlong - haveLong;
			dy = jlat  - haveLat;
			if (dx != 0 || dy != 0) {
			    strokecount++;
			    segsmall=0;
			    if ( abs(dx) > MAX8x  || abs(dy) > MAX8y) {
/* Encode this transition as a pair of 32-bit values. */
				big++;
				bytecount += 8;
				if (segsize+4 > MAXSEG) {
				    fprintf(stderr,"segsize overflow.\n");
				    close(Ofile); fclose(df);
				    return;
				}
				segbuf[segsize++] = (dx >> 16) & ~ SHORTFLAG;
				segbuf[segsize++] = dx & 0xFFFF;
				segbuf[segsize++] = (dy >> 16) & 0xFFFF;
				segbuf[segsize++] = dy & 0xFFFF;
			    } else {
/* Encode this transition as a pair of 8-bit values. */
				small++;
				segsmall=1;
				bytecount += 2;
				if (segsize+1 > MAXSEG) {
				    fprintf(stderr,"segsize overflow.\n");
				    close(Ofile); fclose(df);
				    return;
				}
				segbuf[segsize++] = 0xFFFF & 
				  (SHORTFLAG | 
				   (dx << 8) | 
				   (dy & 0xFF));
			    }
			}
		    }

 if (lsn == 5004620) {
	printf("%3d %2d %2d%c %3d %2d %2d%c %7d %7d ",
	ladg,lam,las,lad,lodg,lom,los,lod,jlat,jlong);
 	if (segsmall) {
	    printf("s: %4d %4d %6o\n",dx,dy,segbuf[segsize-1]);
	} else {
	    printf("l: %4d %4d %6o %6o %6o %6o\n",dx,dy,
		segbuf[segsize-4],segbuf[segsize-3],segbuf[segsize-2],segbuf[segsize-1]);
	}
 }

		    haveLat = jlat;
		    haveLong = jlong;
		}

		sd[segcount].nbytes = bytecount;

		sb.nstrokes = strokecount;
		write(Ofile, &sb, sizeof sb);
		write(Ofile, segbuf, bytecount);
	    }
	    /*  end of each segment */

	    if (segsize > segmax) segmax = segsize;
	}
	/* end of file */
	header.magic	= CBD_MAGIC;
	header.dictaddr	= tell(Ofile);
	header.segcount	= segcount;
	header.segsize	= segcount * sizeof(struct segdict);
	header.segmax	= segmax;

	fprintf(stderr,"Segdict has %d entries, %d bytes, starts at %d.\n	Largest segment has %d entries (%d bytes)\n",
		segcount, header.segsize, header.dictaddr, segmax, 2*segmax);
	write(Ofile, &sd[1], (int) header.segsize);
	lseek(Ofile, 0L, L_SET);
	write(Ofile, &header, sizeof(header));
	close(Ofile);

	fclose(df);
}
