/* 
	Code to handle clipping, both in the lat/long and the x/y domains.
 */
#include "netmap.h"

/* ----------------------------------------------------------------------
   Return true if [x,y] is inside the map region; otherwise false. */

int clip(y,x) double y,x;
{
    if (x<LongMin || x>LongMax) return(FALSE);
    if (y<LatMin || y>LatMax) return(FALSE);
    return(TRUE);
}



/* Clip in the map coordinate system, not in latitude/longitude. Slower,
   but necessary when the map is rotated */
int clipxy(x,y) double x,y;
{
    if (x<ClipXmin || x>ClipXmax) return (FALSE);
    if (y<ClipYmin || y>ClipYmax) return (FALSE);
    return(TRUE);
}
/* ----------------------------------------------------------------------

   If [ox,oy]-->[dx,dy] is a vector that crosses a clipping boundary,
   this function returns the location of the point where it crosses.
   Otherwise it just returns [dx,dy].

   The value returned by the function is 1 if the new point is on a
   horizontal line, and 0 if that point is on a vertical line.
   If the point is solidly inside the clipping region, -1 is returned.

 */

static int quadrants [] = {3,0,2,0,0,0,4,0,1};

atclip (orgx,orgy,dstx,dsty,outx,outy)
double orgx,orgy,dstx,dsty,*outx,*outy;
{
    double slope,intercept,deltax,deltay;
    int quadrant,bdy,retval;

    deltax = dstx-orgx; deltay = dsty-orgy;
    *outx = dstx; *outy = dsty;
    if (clipxy(dstx,dsty)) return (-1);

    retval = 0;
    if (deltax == 0) {
        if (deltay>0) *outy = ClipYmax; 
         else	      *outy = ClipYmin;
	 *outx = orgx;
	 retval = 1;
    } else if (deltay == 0) {
        if (deltax>0) *outx = ClipXmax;
	 else	      *outx = ClipXmin;
    } else {
        slope = deltay/deltax;
	intercept = orgy - orgx*slope;
	quadrant = quadrants[3*(1+sign(deltay)) + (1+sign(deltax))];
	switch (quadrant) {
	    case 1: case 2:
	    	if (orgx == ClipXmax) {bdy = 'r'; quadrant=0; break;}
	    case 3: case 4:
	    	if (orgx == ClipXmin) {bdy = 'l'; quadrant=0; break;}
	}
	switch (quadrant) {
	    case 1: if (slope>((ClipYmax-orgy)/(ClipXmax-orgx)))
	    	    bdy = 't'; else bdy = 'r'; break;
	    case 2: if (slope>((ClipYmin-orgy)/(ClipXmax-orgx)))
		    bdy = 'r'; else bdy = 'b'; break;
	    case 3: if (slope>((ClipYmin-orgy)/(ClipXmin-orgx)))
		    bdy = 'b'; else bdy = 'l'; break;
	    case 4: if (slope>((ClipYmax-orgy)/(ClipXmin-orgx)))
		    bdy = 'l'; else bdy = 't';
	}
	switch (bdy) {
	    case 't': *outy = ClipYmax;
	    	      *outx = (ClipYmax-intercept)/slope;
		      retval = 1;
		      break;
	    case 'b': *outy = ClipYmin;
	    	      *outx = (ClipYmin-intercept)/slope;
		      retval = 1;
		      break;
	    case 'l': *outx = ClipXmin;
		      *outy = slope*ClipXmin + intercept; break;
	    case 'r': *outx = ClipXmax;
		      *outy = slope*ClipXmax + intercept; break;
	}
    }
    return(retval);
}
