#include "netmap.h"
/* 
   This file contains the code that is specific to network host
   and path files, though we don't actually draw paths here. Here are
   the routines that read in host files, record host names, and maintain
   the grid used to prevent hosts from being drawn on top of one another.

 */


/* ----------------------------------------------------------------------
   Keep a pigeonhole table of physical map slots used by host circles, so
   we can avoid putting 2 hosts in the same spot. 
 */

int seqno=0;

/* ----------------------------------------------------------------------- */
/*    Read and process one host file.					 */

int	UseGrid,SomeCircles,	/* derived flags */
	Zmax=0;


gethosts(name,WantPaths,WantFlows)
int 	WantPaths,WantFlows;
char	*name;
{
#define	NTOK	2000
#define	TOKENMAX 50
#define LINESIZE 16384
	FILE	*df;
	char	line[LINESIZE],
		tokens[NTOK][TOKENMAX],
		*tptr,
		*HostName = NULL;
	int	i,k,n,
		Zvalue,
		tpib,			/* this point in bounds? */
		hostid,			/* ID number of current host */
		HostRegistered;		/* set when host entered in smtbl */
	double	Hlat,Hlong,
		Collect();

	if (!(df = fopen(name, "r"))) {
	    perror(name);
	    return;
	}

	fprintf(stderr,"Processing name file %s\n",name);
	switch (HostLabels) {
	    case TextCircles:
	    case Circles: UseGrid=TRUE; SomeCircles=TRUE;
	        break;
	    case TrafficCircles: UseGrid=FALSE; SomeCircles=TRUE;
		printf("CityDotFont CityDotFontSize SCF\n");
	        break;
	    case TextDots:
	    case Dots: UseGrid=FALSE;
		printf("CityDotFont CityDotFontSize SCF\n");
		break;
	    case LabelDots: UseGrid=FALSE;
		printf("CityDotFont CityDotFontSize SCF\n");
		break;
	}

	pathTop = NULL;

	while (fgets(line, LINESIZE, df)) {
/* First we unpack the line into an array of string tokens, in pretty much
   the same way that the shell unpacks input lines. */

	    tptr = line; n = 0;
	    while (*tptr != NULL) {
		while (*tptr==' ' || *tptr=='	' || *tptr=='\n') tptr++;
		if (*tptr == NULL) break;
		i=0;
		while ((*tptr != NULL) &&
			(*tptr != ' ')&&(*tptr != '	')&&(*tptr != '\n'))
		    (tokens[n])[i++] = *tptr++;
		(tokens[n])[i]=(char) 0;
		n++;
		if (i >= TOKENMAX) i--;
		if (n >= NTOK) {
		    fprintf(stderr,"%s entry overflows token table. Skipping.\n",
		        &tokens[0][0]);
		    break;
		}
		if (*tptr == NULL) break;
	    }
/* Now loop over the tokens that we just generated and process some of them
   and ignore others of them. We use "N S E W" as postfix operators to
   retroactively look at the lat/long numbers. Because this loop is executed
   so many times, we don't call strcmp; this gives a startling speedup.
*/
	    Zvalue = 0;
	    Hlat=100; Hlong=999; HostRegistered=0;
	    for(k=0; k<n; k++) {
	        if (tokens[k][0]=='H' && tokens[k][1]==0) {
		    HostName=tokens[0];
		    continue;
		}
	        if (tokens[k][0]=='N' && tokens[k][1]==0) {
		    Hlat=Collect(tokens,k);
		    continue;
		}
	        if (tokens[k][0]=='S' && tokens[k][1]==0) {
		    Hlat= -Collect(tokens,k);
		    continue;
		}
	        if (tokens[k][0]=='E' && tokens[k][1]==0) {
		    Hlong=Collect(tokens,k);
		    continue;
		}
	        if (tokens[k][0]=='W' && tokens[k][1]==0) {
		    Hlong= -Collect(tokens,k);
		    continue;
		}
		if (WantFlows && (tokens[k][0]=='Z' && tokens[k][1]==0)) {
		    Zvalue = atoi(tokens[k-1]);
		    if (Zvalue > Zmax) Zmax = Zvalue;
		}

/* postfix token U means that this is a path from host "hostid" to the  one
   named just before the U. */
		if (WantPaths && (tokens[k][0]=='U' && tokens[k][1]==0)) {
		    char *linknode;
		    int dsthostid;
		    struct hpath *jpath,*kpath;

		    linknode=tokens[k-1];

		    if (!HostRegistered) {
		        if (DebugLevel > 5) {
			    fprintf(stderr,"Registering %s at [%f %f] (%s ",HostName,Hlat,Hlong,prll(Hlat,1,0));
			    fprintf(stderr,"%s)\n",prll(Hlong,0,0));
			}
			if (Hlat == 0 || Hlong == 0) goto nexthost;
		        hostid=reghost(HostName,Hlat,Hlong);
			if (hostid < 0) goto nexthost;
		        HostRegistered++;
			tpib = htable[hostid].inbounds;
			if (htable[hostid].drawn == DrawDefault) {
			    if (DebugLevel > 8) fprintf(stderr,"drawing %s as %d\n",HostName,HostLabels);
			    if (tpib) switch (HostLabels) {
				case TextCircles: htable[hostid].drawn = DrawTextCircle; break;
				case Circles: htable[hostid].drawn = DrawCircle; break;
				case TextTraffic: htable[hostid].drawn = DecideTextTraffic; break;
				case TrafficCircles: htable[hostid].drawn = DecideByTraffic; break;
				case Dots: htable[hostid].drawn = DrawDot; break;
				case TextDots: htable[hostid].drawn = DrawTextDot; break;
			    }
			}
		    }
 
		    dsthostid = reghost(linknode,99.0,999.0);

/* Throw this away if it is a flow line and there is not enough flow over it */
		    if (Zvalue > 0 && Zvalue < MapDetail)
		        goto nextpath;

/* Credit the hosts with this traffic. Note that the sense of the paths is reversed
   from our "origin" and "destination" nomenclature.  */
		   htable[hostid].inTraf += Zvalue;
		   htable[dsthostid].outTraf += Zvalue;

/* See if there is already a path between these two nodes. If there is, then
   don't enter it again, but credit the existing path with any traffic that
   might be on this one. */
		    kpath = htable[hostid].paths;
		    while (kpath != NULL) {
			if ((kpath->oindex == hostid)
			 &&(kpath->dindex == dsthostid)) {
			    kpath->pwidth += Zvalue;
			    goto nextpath;
			}
			kpath = kpath->next;
		    }

/* Don't credit the reverse path--there can be traffic in both directions. */
		    if (!Zvalue) {
			kpath = htable[dsthostid].paths;
			while (kpath != NULL) {
			    if ((kpath->oindex == dsthostid)
			     &&(kpath->dindex == hostid)) {
				goto nextpath;
			    }
			    kpath = kpath->next;
			}
		    }

		    RawPathCount++;
/* link path entry into symbol table entry for origin host */
		    jpath = (struct hpath *) malloc(sizeof(struct hpath));
		    jpath->next = 0;
		    jpath->oindex = hostid;
		    jpath->dindex = dsthostid;
		    jpath->pcolor = LineColor;
		    jpath->passno = NumLayers;
		    if (Zvalue != 0) {
		        jpath->pwidth = Zvalue;
		    } else
		        jpath->pwidth = LineWidth;
		    if (htable[hostid].paths == (struct hpath *) 0) 
		        htable[hostid].paths = jpath;
		    else {
		        kpath = htable[hostid].paths;
			while (kpath != NULL) {
			    if (kpath->next == (struct hpath *) 0) {
			        kpath->next = jpath;
				kpath = (struct hpath *) 0;
			    } else
			        kpath = kpath->next;
			}
		    }
   nextpath:	;
		}
	    }
	    if (!HostRegistered) {
	        if (DebugLevel > 5) {
		    fprintf(stderr,"Registering %s at [%f %f] (%s ",HostName,Hlat,Hlong,prll(Hlat,1,0));
		    fprintf(stderr,"%s)\n",prll(Hlong,0,0));
		}
		if (Hlat == 0 || Hlong == 0) goto nexthost;
	        hostid=reghost(HostName,Hlat,Hlong);
		if (hostid < 0) goto nexthost;
	        HostRegistered++;
	        tpib = htable[hostid].inbounds;
		if (htable[hostid].drawn == DrawDefault) {
		    if (DebugLevel > 8) fprintf(stderr,"drawing %s as %d\n",HostName,HostLabels);
		    if (tpib) switch (HostLabels) {
			case TextCircles: htable[hostid].drawn = DrawTextCircle; break;
			case Circles: htable[hostid].drawn = DrawCircle; break;
			case TextTraffic: htable[hostid].drawn = DecideTextTraffic; break;
			case TrafficCircles: htable[hostid].drawn = DecideByTraffic; break;
			case Dots: htable[hostid].drawn = DrawDot; break;
			case TextDots: htable[hostid].drawn = DrawTextDot; break;
		    }
		}
	    }
/* If the current node type is Circles, then mark the "drawn" field so that a
   circle will be drawn. Otherwise it will get drawn as a dot or as nothing,
   depending on later options. */
 
	    
nexthost: ;
    }

	NumHosts = seqno;
	fclose(df);
	return;
}

/* ----------------------------------------------------------------------- */

/* "Register" a host: make a host table entry for it, give it a host number,
   and find a display slot for it.  Returns the positive hostid, or else
   an error value of -1  */

static struct hn *ctops[26][26] = {0};

int reghost(HostName,Hlat,Hlong)
char *HostName;
double Hlat,Hlong;
{
    double Hx,Hy;
    int tpib,hostid;
    char c,c1,c2;
    int i;
    struct hn *jhlink;
    static char h[HNAMELEN];

    if (!HostName)
	return(-1);

/* This code is ugly, but it's mighty fast. Basically what happened is that I
   wrote this program to handle a small number of nodes, and therefore used
   an unsorted array as a symbol table, and I didn't really want to rewrite
   the program to use a real symbol table, so i added a hash thread through
   the existing table, hashing on the first 2 characters in each name. Also I
   unrolled a bunch of loops and replaced a call to "strcmp" by the gross
   hack that you see below. It's fast. */

    c = *HostName;
    for (i=0; i<HNAMELEN; i++) {
	if (c != 0) c = h[i] = HostName[i];
	    else    h[i] = c;
    }
    c1 = h[0]; c2 = h[1];
    if (c1 < 'a' || c1 > 'z') c1 = 0 ; else c1 = c1 - 'a';
    if (c2 < 'a' || c2 > 'z') c2 = 0 ; else c2 = c2 - 'a';
    jhlink = ctops[c1][c2];

    {	register char *th,*j;
	while (jhlink != NULL) {
	    th = jhlink->hname;
	    j = h;
	    if (
		*th++ == *j++
	     && *th++ == *j++
	     && *th++ == *j++
	     && *th++ == *j++
	     && *th++ == *j++
	     && *th++ == *j++
	     && *th++ == *j++
	     && *th++ == *j++
	     && *th++ == *j++
	    ) {hostid = jhlink->hid; goto foundit;}
	    jhlink = jhlink->cnext;
	}
    }

    seqno++;
    NumHosts = seqno;
    if (seqno >= MAXHOSTS) {
	fprintf(stderr,"Host limit of %d reached: please recompile with larger limit.\n",MAXHOSTS);
	exit(3);
    }
    htable[seqno].paths = NULL;
    htable[seqno].hid = seqno;
    htable[seqno].drawn = DrawDot;
    htable[seqno].inTraf = 0;
    htable[seqno].outTraf = 0;
    htable[seqno].cpoints = 0;
    htable[seqno].datacomplete = 0;
    htable[seqno].hstx = 2*ClipXmax;
    htable[seqno].hsty = 2*ClipYmax;
    /* this copies trailing nulls too, which we want:	 */
    for (i=0; i<HNAMELEN; i++) (htable[seqno].hname)[i] = h[i];
    htable[seqno].hname[HNAMELEN] = NULL;
    hostid = seqno;
    htable[seqno].cnext = ctops[c1][c2];
    ctops[c1][c2] = &htable[seqno];

foundit: 
	if (!htable[hostid].datacomplete
	    && (Hlat < 90.0) && (Hlong < 999.0)) {
	project(Hlat,Hlong,&Hx,&Hy);
	tpib = clipxy(Hx,Hy);
	if (BigImage) {
	    if (!tpib) {
		project(Hlat,Hlong+360.0,&Hx,&Hy);
		tpib = clipxy(Hx,Hy);
	    }
	    if (!tpib) {
		project(Hlat,Hlong-360.0,&Hx,&Hy);
		tpib = clipxy(Hx,Hy);
	    }
	}

	htable[hostid].inbounds = tpib ? 1 : 0;
	htable[hostid].hstx = Hx;
	htable[hostid].hsty = Hy;
	htable[hostid].datacomplete = 1;
    }
    return(hostid);
}



/* ----------------------------------------------------------------------- */


/* Collect a latitude and longitude from the tokens ending with tokens[k] */
double Collect(tokens,k) 
    char tokens[][TOKENMAX];
    int k;
{
    int j;
    double Value;
    for (j=k; j>1; j--) {
	if (!(tokens[j-1][0]=='.' || isdigit(tokens[j-1][0]))) break;
    }
    Value=3600.0*atof(tokens[j++]);
    if (isdigit(tokens[j][0]) || tokens[j][0]=='.') {
	Value=Value+60.0*atof(tokens[j++]);
	if (isdigit(tokens[j][0]) || tokens[j][0]=='.') {
	    Value=Value+atof(tokens[j++]);
	}
    }
    return(Value/3600.0);
}

/* ------------------------------------------------------------------------
   Search for a name in our host table. For the moment this is just a linear
   search. If this gets to be a problem, the host table is in alphabetical
   order and we can use a binary search.

   On success, return the integer subscript in the host table of the entry.
   On failure, return -1.

*/

int findtoken(tname)
    char *tname;
{
    int i;
    for (i=1; i<=NumHosts; i++) {
	  if(!strcmp(tname,htable[i].hname))
	  	return(i);
    }
    return(-1);
}
