#include "netmap.h"
#include "cbdmap.h"

/* ----------------------------------------------------------------------
   Read and process one binary map file.
 */


#define	TRUE	1
#define	FALSE	0

#include <sys/file.h>
#include <string.h>

getmap(name)
char	*name;
{
	int	Ifile,
		inbounds,	/* true after we have drawn an inbounds pt */
		tpib,		/* this point in bounds? */
		segcount,idx,idy,
		pass1,pass2,screen,
		segout,
		segbufsize,
		pass,		/* count multiple passes near date line */
		needPass,	/* true if we need another pass */
		hemisphere,	/* for marking lines across dateline */
		olt,oln,	/* previous point from file */
		j,k,jstroke,iseg;
	double	alt,		/* latitude, in degrees */
		aln,		/* longitude, in degrees */
		prevlt,prevln,	/* previous point, before clipping */
		xposn,yposn,	/* saved position */
		dxposn,dyposn,
		tx,ty,wasx,wasy,
		gymin,gymax,gxmin,gxmax,
		xmax,ymax,xmin,ymin,
		xx,yy;		/* map coordinates */
	unsigned BIT32	uidx;

	struct segdict {
	    BIT32	segid;
	    BIT32	maxlat,minlat,maxlong,minlong;
	    BIT32	absaddr;
	    BIT16	nbytes;
	    BIT16	rank;
	} *sd,*sdbuf;

	struct segment {
		BIT32	orgx,orgy;
		BIT32	id;
		BIT16	nstrokes;
		BIT16	dummy;
	} sb;

	BIT16	*segbuf;

	BIT32	
		i32;
	char	fname[128];
	struct cbdhead header;

/* If the name begins with a /, then use it as a file name. If it does not
   begin with /, then plop the library directory name on the front of it and
   ".cbd" on the back of it.
 */
	if (name[0]=='/')
	    strcpy(fname,name);
	else
	    sprintf(fname,"%s/%s.cbd",LIBDIR,name);
	Ifile = open(fname,O_RDONLY,0);
	if (Ifile < 0) {
	    perror(fname);
	    return;
	}
	fprintf(stderr,"Processing map file %s\n",fname);

/* Read in the file header, check it for the correct magic number,
   and learn the address of the segment dictionary */
	read(Ifile, &header, sizeof header);
	if (header.magic != CBD_MAGIC) {
	    fprintf(stderr,"File %s has bad magic number for .cbd file.\n",
	    		   fname);
	    exit(2);
	}
/* allocate space for the segment buffer */
	segbufsize = 2*header.segmax;
	segbuf = (BIT16 *) malloc(50 + segbufsize);
	
/* allocate space for the segment dictionary */
	sdbuf = (struct segdict *) malloc(100+header.segsize);
	sd = sdbuf;
	sd++;

/* Go read in the segment dictionary (it's at the end of the file) */
	lseek(Ifile, header.dictaddr, L_SET);
	j = read(Ifile, sd, header.segsize);
	if (j < header.segsize) {
	    fprintf(stderr,"File %s: segment dictionary expecting %d got %d\n",
		    fname, header.segsize, j);
	    close(Ifile);
	    free(sdbuf); free(segbuf);
	    return;
	}

	setupLine(LineWidth,LineColor);

/* Now look at each segment and decide if we're going to keep it or not */
	segcount = header.segcount;

	sd = sdbuf;
	segout = 0;
	pass1 = 0; pass2 = 0;
	for (iseg=1; iseg <= segcount; iseg++) {
	    sd++;		/* does this really work? wow! */
#define	LD(x)	(sd->x/3600.0)

/* check to see if this segment is at too fine a detail number */

	    if(sd->rank > MapDetail) continue;

/* if there is a list of included segments, then make sure we are on it */
	    if (numIncludedSegments > 0) {
	       int k,lsn3;
	       lsn3 = ((sd->segid) / 100) % 10000 ;
	       for (k=0; k<numIncludedSegments; k++) {
	           if (lsn3 == IncludedSegments[k]) goto keep;
	       }
	       continue;	/* this segment is not on "keep" list */
	    }
/* Check to see if the segment bounding box is anywhere within the rough
   clipping region. */

keep: ;
	    if (LD(maxlat) < RoughLatMin) continue;
	    if (LD(minlat) > RoughLatMax) continue;

	    if (RoughCrossDateline) {	/* check both intervals */
		if (sign(LD(maxlong)) != sign(LD(minlong))) goto accept;
		if (RoughLongMin < -180.0) {
		    if (LD(maxlong) > RoughLongMin+360.0) goto accept;
		    if (LD(minlong) < RoughLongMax) goto accept;
		} else {
		    if (LD(maxlong) > RoughLongMin) goto accept;
		    if (LD(minlong) < RoughLongMax-360.0) goto accept;
		}
		continue;
	    } else {
		if (LD(maxlong) < RoughLongMin) continue;
		if (LD(minlong) > RoughLongMax) continue;
	    }
accept:	    pass1++;
#ifdef DEBUG
	    if (DebugLevel > 5) {
	        fprintf(stderr,"Seg %d (%d) rough pass; Lat %s--",
		    iseg, sd->segid, prll(LD(minlat),1,0));
		fprintf(stderr,"%s  Long ",prll(LD(maxlat),1,0));
		fprintf(stderr,"%s--",prll(LD(minlong),-1,0));
		fprintf(stderr,"%s\n",prll(LD(maxlong),-1,0));
	    }
#endif DEBUG

/* An exact XY-domain clip requires computing the intersection of two
   arbitrary shapes (the projection of a square in lat-long space can have
   curved edges in XY space). The computation of this intersection is
   remarkably difficult. Specifically, it's more difficult than I want to
   program right now. Also, it's probably more expensive than the computation
   that would be saved by doing it. So we just rely on the rough clip.
 */
	    segout++;
	    lseek(Ifile, sd->absaddr, L_SET);	/* FIXME check result */
	    read(Ifile, &sb, sizeof sb);	/* FIXME check result */
	    if (sd->nbytes > segbufsize) {
	        fprintf(stderr,"%s: segment %d needs %d bytes; buffer limit is %d.\n", fname, iseg, sd->nbytes, segbufsize);
	 	close(Ifile);
		free(sdbuf); free(segbuf);
		return;
	    }

	    read(Ifile, segbuf, sd->nbytes);
/*
  It is possible that this segment will cross the International Date Line.
  If it does, we will want to draw it twice, once for each side, letting each 
  half cross over the line but with a different multiple of 360 added to
  the longitude.
 */
	    needPass = 0;
	    for (pass=0; pass<=1; pass++) {

		k = 0;
		screen = FALSE;
		oln = sb.orgx; olt = sb.orgy;
		hemisphere = sign(oln);
		LongBasis = 360.0 * pass;
		if (pass > 0) {
		    oln = oln - hemisphere*360*3600;
		    hemisphere = -hemisphere;
		}

		project(olt/3600.,oln/3600., &xx, &yy);

		tpib=clipxy(xx,yy);
		if (tpib && clipframe) {
		    tpib=clip(olt/3600., oln/3600.);
		}
		inbounds = FALSE;
		prevlt = 333333;
		prevln = 999999;
		gxmax = -1;
		gxmin = 999999;
		gymax = -1;
		gymin = 999999;
    
		if (tpib) {
		    ymax = ymin = yy;
		    xmax = xmin = xx;
		    xposn = xx; yposn = yy;
		    wasx = xx; wasy = yy;
		    screen = TRUE;
#ifdef SEGLABEL
		    printf("gsave %3.1f %3.1f translate CityDotFont 8 SCF ",
			    xx,yy);
		    printf("20 20 MT -20 -20 L 20 -20 MT\n-20 20 L stroke 20 20 MT (%d) show grestore\n",iseg);
#endif SEGLABEL
		    PSPn(xposn,yposn,"MP");
		    inbounds = TRUE;
		}

		for (jstroke=1; jstroke <= sb.nstrokes; jstroke++) {
		    i32 = 0;
		    if (segbuf[k] & SHORTFLAG) {
/* Flag bit on: unpack a 16-bit field into dx and dy */
			i32 = segbuf[k++];
			if ((i32 & SIGNBIT16) == 0) i32 &= ~SHORTFLAG;
			idy = i32 & 0xFF;
			if (idy & 0x80) idy |= ~ 0xFF; /* extend sign */
			idx = i32 >> 8;
			if (idx & 0x40) idx |= ~ 0x7F; /* extend sign */
			aln = (oln + idx) / 3600.0;
			alt = ((double) (olt + idy)) / 3600.0;
		    } else {
/* Flag bit off: take dx and dy from 32-bit fields.  We do this with
   an unsigned 32-bit int so that overzealous compilers don't sign-extend
   us and mess things up. */
			uidx = segbuf[k++];
			if (uidx & SIGNBIT16) uidx |= SHORTFLAG;
			uidx = (uidx << 16) | (segbuf[k++] & 0xFFFF);
			idx = uidx;
			aln = (oln + idx) / 3600.0;
    
			idy = segbuf[k++];
			idy = (idy << 16) | segbuf[k++];
			alt = ((double) (olt + idy)) / 3600.0;
		    }

/* 
   see if this vector has crossed the Dateline, and, if it has, then do 
   the right things. Specifically, we want to convert it to a value that
   is numerically correct for the hemisphere that the segment started
   in, and we want to set a flag requesting a second pass through the
   data.
 */
		    if ((sign(oln) != sign(aln)) &&(abs(idx) > 90.0*3600.0)) {
			needPass++;	/* ask for a second pass */
			oln = oln + idx - 360*3600*sign(aln);
			aln = aln - 360.0 * sign(aln);
			
		    } else {
			oln = oln + idx;
		    }
		    if (abs(aln)>360.0) aln = aln - 360.0 * sign(aln);
		    if (abs(oln)> 360*3600 ) oln = oln - 360*3600*sign(oln);
		    olt = olt + idy;
		    tpib=FALSE;

 /*
    If this segment is in the general vicinity of the dateline, then,
    even if it doesn't cross the dateline, we should repeat it with a
    different hemisphere offset. This will allow nice continuity of
    the continent shapes on maps whose overwall width is greater than
    360 degrees.
  */
		    if (abs(aln) > 90.0) needPass++;
		/*    needPass++;   */

 /* If this point fails the rough-clip test, then don't bother
    looking further. However, rough clip gives false positives, so if
    it passes, we have to look more carefully
  */
		    if ((alt > RoughLatMin) &&
		       (alt < RoughLatMax) &&
		       (aln > RoughLongMin) &&
		       (aln < RoughLongMax)) { 
			project(alt,aln,&xx,&yy);
			tpib=clipxy(xx,yy);
			if (tpib && clipframe) {
			    tpib=clip(alt,aln);
			}
		    } else if (inbounds) {
			project(alt,aln,&xx,&yy);
		    }
		    if (tpib) {
			screen = TRUE;
			dxposn = xx - xposn;
			dyposn = yy - yposn;
			if (yy < ymin) ymin = yy;
			if (yy > ymax) ymax = yy;
			if (xx < xmin) xmin = xx;
			if (xx > xmax) xmax = xx;
 /*
     If this point is inside the clipping region but the previous point
     was not, then we need to compute the coordinates of the place where
     the line segment crossed the clipping region, and pretend that
     the segment started there. This ends up computing a false value
     for the "previous position" of the cursor.
  */

			if (!inbounds) {
			    project(prevlt,prevln,&xposn,&yposn);
			    atclip(xposn,yposn,xx,yy,&tx,&ty);
			    xposn = tx; yposn = ty;
			    PSPn(xposn,yposn,"MP");
			    dxposn = xx - xposn;
			    dyposn = yy - yposn;
			    inbounds=TRUE;
			}
 
			LSP(dxposn,dyposn);
			wasx = xx; wasy = yy;
		    } else {
 /*
    If we used to be inside the clipping region but now we aren't, then we
    have to do the reverse of the above-mentioned computation, namely, find
    the place where this line segment exits from the clipping area and then
    draw the line to exactly that point, but no further.
  */
			if (inbounds) {
			    atclip(wasx,wasy,xx,yy,&tx,&ty);
			    dxposn=tx-xposn; dyposn=ty-yposn;
			    FlushLSP(TRUE);
			    PSP(dxposn,dyposn,"MB");
			    inbounds=FALSE;
			}
			xposn = xx; yposn = yy;
		    }
    
		    prevlt = alt; prevln = aln;
		}
		if (gxmax < xmax) gxmax = xmax;
		if (gxmin > xmin) gxmin = xmin;
		if (gymax < ymax) gymax = ymax;
		if (gymin > ymin) gymin = ymin;
    
		if (screen) pass2++;
		if (inbounds) {
		    FlushLSP(TRUE);
		    printf("ST\n");
		}
		inbounds = FALSE;
		if (!needPass) break;
	    }
	}
	close(Ifile);
	if (DebugLevel>0)
	    fprintf(stderr,"%d segments; %d passed rough clip; %d in bounds.\n",
		segcount,pass1,pass2);
	free(sdbuf);
	free(segbuf);
}
