#include "netmap.h"

static double WantX,WantY,HaveX,HaveY,LatentX,LatentY;
int	pathsize,ntpl,LatentData=0;
int  CumStrokeCount=0,CumRawStrokeCount=0;

/* ---------------------------------------------------------------------- */
/* Move or stroke to a point, and initialize the optimizer */

PSP(x,y,F) double x,y; char *F;
{
    int ix,iy;

    ix = trunc(x); iy = trunc(y);
    if (Exact) {
        printf("%4.2f %4.2f %s\n",x,y,F);
    } else {
        printf("%d %d %s\n",ix,iy,F);
    }
    CacheActive=FALSE;
    LatentData=0;
    pathsize = 0;
    HaveX=HaveY=0;
}

/* ---------------------------------------------------------------------- */
/*  Same as PSP, but doesn't write a newline.  */

PSPn(x,y,F) double x,y; char *F;
{
    int ix,iy;

    ix = trunc(x); iy = trunc(y);
    if (Exact) {
        printf("%4.2f %4.2f %s ",x,y,F);
    } else {
        printf("%d %d %s ",ix,iy,F);
    }
    CacheActive=FALSE;
    LatentData=0;
    pathsize = 0;
    HaveX=HaveY=0;
}

/* ---------------------------------------------------------------------- */


/* Write out rounded coordinates, but not any opcode. */
PP(x,y) double x,y;
{
    int ix,iy;

    ix = x; iy = y;
    printf("%d %d ",ix,iy);
}

/* ----------------------------------------------------------------------

   Srtoke optimizer. This routine is called by the graphics code when it
   wants to draw a line segment. We keep careful watch on the
   segments being drawn, and only draw ones that will be visible on
   the device at the requested resolution.
 */

LSP(x,y) double x,y;
{   int ix,iy;
    double oldslope,newslope;
    int olddx,olddy,newdx,newdy,Dsquared,DDsquared;

/* CacheActive TRUE means that LSP has been called and that there is state in
   the cache, but that no commands have been output. Each time commands are
   output, CacheActive is set FALSE, and each time we absorb position changes
   without writing anything, it becomes TRUE. 

*/

    CumRawStrokeCount++;
    ix = trunc(x); iy = trunc(y);
    olddx=trunc(WantX-HaveX); olddy=trunc(WantY-HaveY);
    newdx = ix-HaveX; newdy = iy-HaveY;
    if (CacheActive) {

/* Check for no motion at all */
	Dsquared = (newdx*newdx + newdy*newdy);
	if (Dsquared <= MinimumStrokeSize) goto xit2;

/* Check for continued vertical and horizontal motion in same direction */
	if (olddx==0 && newdx==0 && (sign(newdy)==sign(olddy))) goto xit;
	if (olddy==0 && newdy==0 && (sign(newdx)==sign(olddx))) goto xit;

/* Check for continued motion along same slope */
	if ((olddx != 0) && (newdx != 0)) {
	    oldslope=((float)olddy)/((float)olddx);
	    newslope=((float)newdy)/((float)newdx);
	    if ((oldslope==newslope) && (sign(olddx)==sign(newdx))) goto xit;
	}

#define ENDCUT	/* enable more intense optimizing of wavy lines */
#ifdef ENDCUT
/* For the moment it would seem that we have to flush the path, but before we
   do, check to see whether the new path is different enough from the old one
   that it is really necessary. */
	newdx = trunc(ix - WantX);
	newdy = trunc(iy - WantY);
	DDsquared = (newdx*newdx + newdy*newdy);
	if (DDsquared < MinimumStrokeSize) {
	    LatentX = ix; LatentY = iy; LatentData = 1;
	    goto xit3;
	}
#endif ENDCUT

/* Alas, the direction has changed consequentially: output the stroke */
	FlushLSP(TRUE);
    }

xit:
xit2:
    WantX=x; WantY=y;
xit3:
    CacheActive = TRUE;
    return;

}

/* ---------------------------------------------------------------------- */
/* This is a table of 1-character PostScript functions, which are defined
   in the profile (netmap.pro). They generate relative cursor motion according
   to their position relative to the center in this matrix. For example, "r"
   moves one space to the right and "l" moves one space to the left.
 */
    static char *ntbl[7] = {
       	"MxwvXXZ",
	"ysmUiJX",
	"zqhuenV",
	"AWlXrET",
	"BpgdfoS",
	"CtkDjKQ",
	"FGHINOP"  }
    ;

/* ---------------------------------------------------------------------- */
/* Dump the accumulated relative motion to a path command */
FlushLSP(KeepLatent)
int KeepLatent;
{
    int dx,dy,row,col,abbrev;
    if (KeepLatent && LatentData) {
        WantX = LatentX; WantY = LatentY;
	LatentData = 0;
    }

    dx = trunc(WantX-HaveX); dy = trunc(WantY-HaveY);
    if ((dx != 0) || (dy != 0)) {
	CumStrokeCount++;
	pathsize++; ntpl++;
	row=3-dy; col=3+dx;
	if (row >= 0 && row <= 6 && col >= 0 && col <= 6) {
	    abbrev=(ntbl[row])[col];
	    if (abbrev != 'X') {
	        printf("%c ",abbrev);
		goto out;
	    }
	}
	if (dx == 0) {printf("%d Y ",dy); ntpl+= 2; goto out;}
	if (dy == 0) {printf("%d X ",dx); ntpl+= 2; goto out;}
	printf("%d %d R ",dx,dy); ntpl+= 5;
    }
out: CacheActive=FALSE;
/* This next step is very crucial to the accuracy of the paths drawn. Because
   we are generating only integer coordinates on the motion commands, an
   error will accumulate. By adding dx and dy to HaveX and HaveY, rather than
   just setting them equal to WantX and WantY, we track the actual cursor
   position including rounding (but not including floating point error in
   this program).
 */

    HaveX=HaveX+dx;
    HaveY=HaveY+dy;

    if (pathsize > 200) {
        printf("MI ");
	pathsize = 0;
    }
    if (ntpl>40) {printf("\n"); ntpl=0;}
}

/* ---------------------------------------------------------------------- */

/* 
   This function decides how wide to make a path line. and what color to
   make it. If the variable Zmax is zero, that means that path widths will
   be determined by options to the program; if Zmax is nonzero, that means
   that path widths will be determiend by data in each path record. If
   "pc", path color, has a nonzero value, then we set a color before 
   drawing this path.
 */
setupLine(pw,pc)
int pw,pc;
{
    double Pval;
    static int CurrentPW = -99;
    static int CurrentPC = 0;
    int ThisPW;

    if ((pc != CurrentPC) && (pc > 0) && (pc <= NumColors)) {
	printf(" %d SLC ",pc);
	CurrentPC = pc;
    }

    ThisPW=pw;
    if (Zmax == 0) {
	if (pw > 4 || pw < -8) return(0);
	if (pw > 0) {
	    if (ThisPW != CurrentPW) printf("%s SLW ",WidthNames[pw-1]);
	} else if (pw >= -4) {
	    if (ThisPW != CurrentPW) printf("%s SDW ",WidthNames[-(1+pw)]);
	} else if (pw >= -8) {
	    if (ThisPW != CurrentPW) printf("%s SFW ",WidthNames[-(5+pw)]);
	}
	CurrentPW = ThisPW;
	return(1);
    } else {
	pw = abs(pw);
	Pval = ((double) pw) / Zhostmax;
        if (ThisPW != CurrentPW) printf("%3.2f SPW ",Pval);
	CurrentPW = ThisPW;
	return(TrafficLW*Pval);
    }
}
