/* 
   This file contains the function "init", which is called once from main
   after all of the options have been scanned. "init" does all of the various
   initialization that needs to be done.

   Brian K. Reid, DEC Western Research, 1986
 */

#include "netmap.h"
#include <time.h>
#include <pwd.h>	   
#include <strings.h>	   

char *rindex();

	char	profilename[100]="";
	double ScaleFactor;
init (argv) 
char **argv;
{
	FILE	*hf;
	char	line[200];
	int	i,zoom,mapsize,lineCount;
	int	EPSFheadLineCount;
	int	ScanColors=0;
	double	ymin,ymax,xmin,xmax,
		Xscf,Yscf,scf,	/* scale factors */
		llx,
		lly,
		xx,yy;		/* scratch variables */
	char *p;
	static char pwname[100], pwgecos[100];

/* Go get the date and time and store it in a global string */
	{
	    long lt;
	    char *ts;
	    time(&lt);
	    ts=ctime(&lt);
	    ts[24]=NULL;
	    strcpy(sdat,ts);
	}

/* Go get the passwd entry that shows who is running the program, and
   copy information that we care about into static strings. */
	{
	    struct passwd *jpw,*getpwuid();
	    char *cp,*strchr();

	    jpw = getpwuid(getuid());
	    strcpy(pwname, jpw->pw_name);
	    strcpy(pwgecos,jpw->pw_gecos);
     /* if there is junk after the name, don't use it */
	    if ((cp=strchr(pwgecos,',')) != (char *) NULL) {
	        *cp = (char) NULL;
	    }
	}

/* Copy the PostScript profile to the output, stripping comment lines.
   As we do this, we copy in certain configuration data from the profile.
 */
	if (CannedMap) {
	    printf("%%!\n%% %s %s\n",ProgID,sdat);
	}
	
	p = rindex(argv[0], '/');
	if (p)	p++;
	else	p = argv[0];
	if (profilename[0]==NULL)
	    sprintf(profilename,"%s/%s.pro",LIBDIR,p);
	if (!(hf = fopen(profilename, "r"))) {
	    fprintf(stderr,"Profile file missing or unreadable.\n");
	    perror(profilename);
	    exit(1);
	}
 fprintf(stderr,"Reading %s\n",profilename);

	EPSFheadLineCount = 0;
	lineCount = 0;
	while (fgets(line, 100, hf)) {
	    if (!CannedMap) {
		lineCount++;
		if (!strncmp(line,"%!PS-Adobe",10)) {
		    if (!NoProlog) fputs(line,stdout);
		    EPSFheadLineCount++;
		}
		if (!strncmp(line,"%%BoundingBox:",14)) {
		    if (!NoProlog) fputs(line,stdout);
		    EPSFheadLineCount++;
		}
		if (!strncmp(line,"%%DocumentFonts:",16)) {
		    if (!NoProlog) fputs(line,stdout);
		    EPSFheadLineCount++;
		}
		if (!strcmp(line,"%%*\n")) {
		    EPSFheadLineCount++;
		    if (!NoProlog) {
			printf("%%%%Pages: 1\n");
			if (MapTitle != NULL) printf("%%%%Title: %s \n",MapTitle);
			printf("%%%%Creator: %s (%s) with %s\n",pwname,pwgecos,ProgID);
			printf("%%%%CreationDate: %s\n",sdat);
		    }
		}
		if (!strcmp(line,"%%EndComments\n")) {
		    EPSFheadLineCount++;
		    if (!NoProlog) fputs(line,stdout);
		}
		
		if ((line[0] == '%') || (line[0] == '\n')) continue;
		if (!NoProlog) fputs(line,stdout);
	    }


	    if ((line[0] == '%') || (line[0] == '\n')) continue;
	    if (!strncmp(line,"/Pwid ",6))
		getPSarg(line,&PageWidth,PageWidth);
	    if (!strncmp(line,"/Pht ",5))
		getPSarg(line,&PageHeight,PageHeight);
	    if (!strncmp(line,"/Margin ",8))
		getPSarg(line,&PageMargin,PageMargin);
	    if (!strncmp(line,"/CaptionMargin ",15))
		getPSarg(line,&CaptionMargin,CaptionMargin);
	    if (!strncmp(line,"/Clw ",5))
		getPSarg(line,&TrafficLW,TrafficLW);

	    if (ScanColors) {
	    	char colorname[100];
		int nscan;

		if (!strcmp(line,"end\n")) {
		    ScanColors = FALSE;
		    continue;
		}
	        nscan=sscanf(line,"/%s",colorname);
		if (nscan == 0) {
		    fprintf(stderr,"%s: %s\nThe profile (%s) contains an illegal color definition line.\n",profilename,line,profilename);
		    continue;
		}
		NumColors++;
		colorNames[NumColors] = (char *) malloc(1+strlen(colorname));
		strcpy(colorNames[NumColors],colorname);
	    }
	    if (!strcmp(line,"Colors begin\012")) ScanColors = TRUE;
	}
	fclose(hf);

/* delicate hack. This should be done better, but I'm in a hurry.
   check to make sure we found all of the EPSF header lines in
   the profile. Better to keep a table, by name, of the ones that
   we found, so the error message can say which ones are mising.  */

	if ((!CannedMap) && EPSFheadLineCount != 5) {
	    fprintf(stderr,"Fatal error: %s.pro has incorrect EPSF headers (found only %d)\n",
	        p,EPSFheadLineCount);
		exit(1);
	}

	if (!CannedMap) printf("gsave\nINIT\n");

/* Now that we know the page size, allocate arrays whose size depends on the
   page size.	 */
	if (PageWidth <= 0) {
	    fprintf(stderr,"Fatal error: %s.pro does not set /Pwid\n",p);
	    exit(1);
	}
	if (PageHeight <= 0) {
	    fprintf(stderr,"Fatal error: %s.pro does not set /Pht\n",p);
	    exit(1);
	}
	if (PageMargin <= 0) {
	    fprintf(stderr,"Fatal error: %s.pro does not set /Margin\n",p);
	    exit(1);
	}

	if (CaptionMargin <= 0) {
	    fprintf(stderr,"Fatal error: %s.pro does not set /CaptionMargin\n",p);
	    exit(1);
	}
	ImageWidth = PageWidth - 2*PageMargin;
	ImageHeight = PageHeight - (2*PageMargin+CaptionMargin);

	XSlots = (1+(ImageWidth/SlotSize));
	YSlots = (1+(ImageHeight/SlotSize));
	if (DebugLevel > 0) fprintf(stderr,"DebugLevel=%d\n",DebugLevel);
	if (DebugLevel > 0)
	    fprintf(stderr,"Paper size %3.1f in. wide by %3.1f in. high\n",
		PageWidth/72.0,PageHeight/72.0);
	mapsize = (4+XSlots)*(4+YSlots)*sizeof(short);
	cmap = (short *)malloc(mapsize);
	bzero(cmap, mapsize);
	if (DebugLevel > 1)
	     fprintf(stderr,"Slot map size is %d bytes (%d by %d)\n",mapsize,XSlots,YSlots);

   
	initProj();		/* initialize the projection */

	if (LongWidth<=0) {
	    fprintf(stderr,"Error in -w option. LongWidth=%d.\n",LongWidth);
	    exit(2);
	}

	identProj();		/* tell it who we are */

	LatMin=LatCenter-(LatHeight/2.0);
	LatMax=LatCenter+(LatHeight/2.0);
	if (Projection == Mercator || Projection == Gall
	  || Projection == CentralCyl) {
	    if (LatMax>75.) LatMax = 75.;
	    if (LatMin<-75.) LatMin = -75;
	    LatHeight = LatMax - LatMin;
	}
 	LongMin=LongCenter-(LongWidth/2.0);
	LongMax=LongCenter+(LongWidth/2.0);
	if ((LongMin != fixlong(LongMin)) || (LongMax != fixlong(LongMax)))
	    CrossDateline = TRUE;


	MapScale = MapScaleBase/(LongWidth<LatHeight ? LongWidth : LatHeight);
	if (LongWidth > 100.0) BigImage = TRUE;

/* 
   Figure out the scale factor for the map. This is a real hack.
   We make 2 loops through the initialization code. The first time
   through we use a completely arbitrary scale factor, and do all of the
   scaling and see how it comes out. Then we change the numerator of the
   projection fraction by the ratio of what we got to what we wanted, and
   then swing through the loop a second time. I'm sure that there is a
   closed-form analytic solution to this problem, but I'm equally sure
   that it's hard, and this loop executes only once each time the program
   starts up...
 */
	for (zoom=0; zoom <= 1; zoom++) {
	    static double xs[7],ys[7]; /* compiler croaks unless this static */
	    double Xsymd,Ysymd,ScaleAdjust;
	    int RDsave;

	    CenterX = CenterY = 0.0; rotate = 0.0;
	    RDsave = RotDegrees; RotDegrees = 0;
	    project(LatCenter,LongCenter,&xx,&yy);
	    CenterX = xx; CenterY = yy;
	    RotDegrees = RDsave;
	    rotate=rads(((double) RotDegrees));
	    if (RotDegrees==90 || RotDegrees==180 || RotDegrees== -90) {
		CaptionAngle = -RotDegrees;
	    }

	    project(LatMin,LongMin,&xs[0],&ys[0]);	/* ll */
	    project(LatMin,LongMax,&xs[1],&ys[1]);	/* lr */
	    project(LatMax,LongMax,&xs[2],&ys[2]);	/* ur */
	    project(LatMax,LongMin,&xs[3],&ys[3]);	/* ul */
	    project(LatMin,LongCenter,&xs[4],&ys[4]);	/* bc */
	    project(LatMax,LongCenter,&xs[5],&ys[5]);	/* tc */
	    xmax = xmin = xs[0]; ymax = ymin = ys[0];
	    for (i=1; i<=5; i++) {
		if (xmax<xs[i]) xmax=xs[i];
		if (xmin>xs[i]) xmin=xs[i];
		if (ymax<ys[i]) ymax=ys[i];
		if (ymin>ys[i]) ymin=ys[i];
	    }
/*	    ulx=xs[3]; uly=ys[3];    urx=xs[2]; ury=ys[2]; */
/*	    lrx=xs[1]; lry=ys[1]; */

	    llx=xs[0];
	    lly=ys[0]; 

/* Compute scale factor based on symmetric bounding box (rotated projection
   is sometimes L/R asymmetric)					 */
	    Xsymd = (-xmin)>xmax ? -2.0*xmin : 2.0*xmax;
	    Ysymd = (-ymin)>ymax ? -2.0*ymin : 2.0*ymax;
	    Xscf = ((double)ImageWidth)/Xsymd;
	    Yscf = ((double)ImageHeight)/Ysymd;
	    scf = Xscf < Yscf ? Xscf : Yscf;

/* This clipping region corresponds to the border at the edge of the page */
	    ClipXmax=(((double)ImageWidth)/2)/scf;
	    ClipXmin= -ClipXmax;
	    ClipYmax=(((double)ImageHeight)/2)/scf;
	    ClipYmin= -ClipYmax;
	    if (clipframe) {
	        ClipYmax=ymax;
		ClipYmin=ymin;
	    }

	    ScaleAdjust = (DeviceResolution * scf) / 72.0 ;
	    if (zoom==0)
	        MapScale *= ScaleAdjust;
	}
	ScaleFactor = scf;

/* Now that we know the clipping region, we can compute the slot increments */
	Xslotinc = (ClipXmax-ClipXmin)/XSlots;
	Yslotinc = (ClipYmax-ClipYmin)/YSlots;



/* If the map area will not be rectangular in Lat/Long space (i.e. if the -f
   option is specified) then we must get ourselves some extra ammunition to
   make clipping happen faster. */
#ifdef DEBUG
 if (DebugLevel > 0) 
  {
    fprintf(stderr,"Clipping region: %f,%f   to  %f,%f\n",
  	ClipXmin,ClipYmin,ClipXmax,ClipYmax);
  }

#endif DEBUG
	{
	    /* this tickles a C compiler bug unless it's static !!??!! */
	    static double aa[8], bb[8]; 

	    normalizeLongs=0;		/* turn off normalization */
	    unproject(ClipXmax,ClipYmax,&aa[0],&bb[0]);
	    unproject(ClipXmax,ClipYmin,&aa[1],&bb[1]);
	    unproject(ClipXmin,ClipYmax,&aa[2],&bb[2]);
	    unproject(ClipXmin,ClipYmin,&aa[3],&bb[3]);
	    unproject(0.0,     ClipYmax,&aa[4],&bb[4]);
	    unproject(0.0,     ClipYmin,&aa[5],&bb[5]);
	    unproject(ClipXmax,0.0     ,&aa[6],&bb[6]);
	    unproject(ClipXmin,0.0     ,&aa[7],&bb[7]);
	    RoughLongMin=RoughLongMax=bb[0];
	    RoughLatMin=RoughLatMax=aa[0];
	    for (i=1;i<=7;i++) {
	        if (RoughLongMin>bb[i]) RoughLongMin=bb[i];
	        if (RoughLongMax<bb[i]) RoughLongMax=bb[i];
	        if (RoughLatMin>aa[i]) RoughLatMin=aa[i];
	        if (RoughLatMax<aa[i]) RoughLatMax=aa[i];
	    }
	    normalizeLongs=1;		/* back on again */

	    if (RoughLongMin < -180.0 || RoughLongMax > 180.0)
	    	RoughCrossDateline++;
	}
	if (DebugLevel>0) {
	    fprintf(stderr,"Rough clip outline: Lat %s -- ",
	    	prll(RoughLatMin,1,0),RoughLatMin);
	    fprintf(stderr,"%s (%4.2f  %4.2f),  Long ",
	    	prll(RoughLatMax,1,0),RoughLatMin,RoughLatMax);
	    fprintf(stderr,"%s -- ",prll(RoughLongMin,-1,0));
	    fprintf(stderr,"%s (%4.2f  %4.2f)\n",
	    	prll(RoughLongMax,-1,0),RoughLongMin,RoughLongMax);
	}
	if (CannedMap) {
	    printf("%%%%CMAP %f %f %f %f %f %d %d %d %f %f %d %d %d %d %d\n",
		LatCenter,LongCenter,LatHeight,LongWidth,MapScale,
		StrokeLimit,RotDegrees,Projection,Lamb1,Lamb2,
		DeviceResolution,PageHeight,PageWidth,PageMargin,
		CaptionMargin);
	    printf("%% %s\n",MapTitle);
	    return;
	}
        printf("/CxD %d 2 div def\n",SlotSize);

/* Put legend in lower right corner of the map */
	if (OmitDetails == 0) {
	    printf("LGcolor\n");
	    printf("(DECWRL %s by Brian Reid at %s) 1 MLB\n",ProgID,sdat);
	    printf("(");
	    identify();
	    printf(",  Map center: [%s, ", prll(LatCenter,1,1));
	    printf("%s]) 2 MLB\n", prll(LongCenter,-1,1));
	    printf("(Image resolution %d/in., stroke limit %d pixels",
		DeviceResolution, StrokeLimit);
	    if (RotDegrees % 90) printf(", rotated %d\\312",RotDegrees);
	    printf(") 3 MLB\n");
	}

	PSP(ImageWidth/2.0,ImageHeight/2.0,"translate");
	printf ("%f dup dup scale /scf exch def\n", scf);
	printf ("/DAR [ DAR {DV} forall ] def\n");
	printf ("/DSR [ DSR {DV} forall ] def\n");
	printf ("/AFont ArrowFont 10 DV scalefont def\n");

	if (showframe) {
	    double Px,Py;
	    Exact = TRUE;
	    printf("newpath FGcolor\n");
	    PSP(llx, lly, "MP"); Px=llx; Py=lly;
#ifdef DEBUG
  if (DebugLevel > 0) {
 	fprintf(stderr,"Frame %s ",prll(LatMin,1,0));
	fprintf(stderr," %s ",prll(LongMin,-1,0));
	fprintf(stderr," to %s ",prll(LatMax,1,0));
	fprintf(stderr," %s\n",prll(LongMax,-1,0));
	
}
#endif DEBUG
	    for(i=1; i<=20; i++) {
		project(LatMin,LongMin+((i*LongWidth)/20.0),&xx,&yy);
		PSP(xx-Px,yy-Py,"R"); Px=xx; Py=yy;
	    }
	    for(i=1; i<=20; i++) {
		project(LatMin+((i*LatHeight)/20.0),LongMax,&xx,&yy);
		PSP(xx-Px,yy-Py,"R"); Px=xx; Py=yy;
	    }
	    for(i=20; i>=1; i--) {
		project(LatMax,LongMin+((i*LongWidth)/20.0),&xx,&yy);
		PSP(xx-Px,yy-Py,"R"); Px=xx; Py=yy;
	    }
	    for(i=20; i>=1; i--) {
		project(LatMin+((i*LatHeight)/20.0),LongMin,&xx,&yy);
		PSP(xx-Px,yy-Py,"R"); Px=xx; Py=yy;
	    }
	    printf("closepath ST\n");
	    Exact = FALSE;
	}
	if (gridincr > 0) {
	    int OLW;
	    OLW = LineWidth;
	    LineWidth = gridStyle;
	    printf("FGcolor\n"); drawgrids();
	    LineWidth = OLW;
	}
	printf("%%%%Endinit\nMFcolor\n");
}



getPSarg(line,var,xval)
char *line;
int *var,xval;
{
    char dname[30],cname[30];
    int vval;
    sscanf(line,"%s%d%s", dname, &vval, cname);
    if (UsingCanned && (vval != xval) && (xval != 0)) {
        fprintf(stderr,"%s: %s\nCanned map file uses value of %d.\nThis value in profile is different and is being ignored.\n",profilename,line,xval);
	*var = xval;
	return;
    }
    *var = vval;
    if (!strcmp(cname,"def")) return;
    fprintf(stderr,"format error in %s: %s", profilename, line);
    exit(2);
}
