/* 
   This program reads text lines from the CIA World Bank II geographic
   database and produces a PostScript file that is a map. It lets you choose
   a Mercator or Lambert projection, lets you choose the level of detail
   selection from the CIA database, and can also read in a file of node names
   and locations and mark the node locations with dots or names or numbers.

	Brian K. Reid, DEC Western Research, April 1986
	Last updated v2.1a May 1993


 */

#include "netmap.h"

    char	ProgID[] = "netmap-2.1";

    int		BiDirectional=0,
    		BigImage=0,	/* true if image large enough to wrap */
    		CacheActive,
		CannedMap=0,
		DebugLevel=0,
		UsingCanned=0,
		NoProlog=0,
		ErrorMessages=0,
		Projection=0,	/* what kind of map projection ? */
		HostLabels=Circles,  /* what kind of host labels ? */
		IncludedSegments[500],
		numIncludedSegments=0,
		Exact,		/* True ==> we want exact positioning */
		MapDetail=3,	/* how much to include in the map */
		DeviceResolution=300, /* dots per inch */
		MinimumStrokeSize=1, /* min vector length in LSP */
		StrokeLimit,	/* same value not squared */
		PageWidth,	/* in points. value set from profile */
		PageHeight,	/* as it is being copied to output */
		PageMargin,	/* how much margin around the edge? */
		ImageWidth,	/* printable area */
		SlotSize=11,	/* diameter in points of node circle */
		gridStyle,
		ImageHeight,
		NumHosts,
		OmitDetails=0;	/* true to omit caption details */
		NHIB,		/* Number of hosts in bounds */
		NumPaths,	/* path count after clipping */
		NumColors=0,	/* number of defined color names */
		NumLayers=0,	/* number of color priority layers */
		RawPathCount=0,	/* path count before clipping */
		RotDegrees=0,
		LineWidth=1,	/* how dark to make the lines */
		LineColor=0,	/* what color are the lines */
		CrossDateline=0,
		showframe=0;
		clipframe=0;
		normalizeLongs=1; /* see "unproject() */
    double	Lamb1,Lamb2,LatCenter,LongCenter,LatMin,LatMax,
	/* the area covered by the map */
		LongMin,LongMax,LatHeight,LongWidth,
	/* a multiple of 360 degrees used for wraparound projection */
		LongBasis,
	/* coordinates for clipping in projected domain */
		ClipXmax,ClipXmin,ClipYmax,ClipYmin,
	/* rough clipping in lat/long domain before XY clip: */
		RoughLongMin,RoughLongMax,RoughLatMin,RoughLatMax,
		RoughCrossDateline,

		gridincr=0.0,
		Cutoff=0.0,
		fixlong(),
		CenterY,CenterX,
		CaptionAngle=0.0,
		MapScaleBase=100000.0,
		rotate=0,
		MapScale;

    char	*MapTitle=NULL,
	        *MapSubtitle=NULL,
		*colorNames[50],
		sdat[28];	/* string date and time */

    char	WidthNames[5][4] = {"Nlw", "Mlw", "Wlw", "Xlw", "Zlw"};
    extern int CumStrokeCount,CumRawStrokeCount;
/* ---------------------------------------------------------------------- */


main(argc, argv)
int	argc;
char	**argv;
{
	extern int optind;
	extern char *optarg;
	extern char *strcpy();
	int	c,
		initdone=FALSE,
		j,nfields,
		IncludeSeg,
		badopts=0;
	char	*tc,
		arg1[100],arg2[100],	/* scratch for scanf */
		colorName[100];
		

	fprintf(stderr,"\n%s\n",ProgID);
	while ((c=getopt(argc, argv,
			"cCd:D:efFg:G:hil:k:mM:nN:Op:P:R:r:s:S:T:t:w:W:yzZ")) != EOF)
	    switch (c) {
/* -c: can. Make a canned PostScript map for later use by -M option. */
		case 'c': initcheck(initdone,c,&badopts);
			  CannedMap = TRUE;
			  break;
/* -C: continue. This map is a continuation and does not need a prolog. */
		case 'C': initcheck(initdone,c,&badopts);
			  NoProlog++;
			  break;
/* -d: detail. Argument is an integer giving map detail to include. */
		case 'd': sscanf(optarg,"%d",&MapDetail);
			  break;
/* -D: debug. Turn on debugging. Argument is debug level. 	 */
		case 'D': sscanf(optarg,"%d",&DebugLevel);
			  break;
/* -e: error messages. Typeset a page with err msgs about the data. */
		case 'e': ErrorMessages++;
		 	  break;

/* -f: show the window frame that has determined scale and placement */
	        case 'f': showframe=1; break;

/* -F: flush all paths to nodes for which no definition has been seen */
		case 'F': flushpaths();
			  break;
/* -g: grid. Arguments in minutes. "-g 60" puts grid lines every degree */
		case 'g': initcheck(initdone,c,&badopts);
			  gridincr=scanlat(optarg);
			  gridStyle=LineWidth;
			  break;
/* -G: diameter of node circles */
		case 'G': initcheck(initdone,c,&badopts);
			  sscanf(optarg,"%d",&SlotSize);
			  break;

/* -h: clip so nothing shows outside of the locator frame */
		case 'h': clipframe=1; break;

/* -i: include segments with this 3-digit lsn prefix */
		case 'i':numIncludedSegments = 0; 
			 for(;optind<argc && argv[optind][0]!='-';optind++) {
			      if (!initdone) init(argv);
			      initdone++;
			      sscanf(argv[optind],"%d",&IncludeSeg);
			      IncludedSegments[numIncludedSegments++] = IncludeSeg;
			  }
			  break;

/* -k (kolor): specify the name of a color for upcoming lines */
		case 'k': if (!initdone) init(argv);
			  initdone++;
			  strcpy(colorName,optarg);

			  for (j=1; j<=NumColors; j++) {
			      if (!strcmp(colorName,colorNames[j])) {
			          LineColor = j;
				  NumLayers++;
				  goto gotcolor;
			      }
			  }
			  fprintf(stderr,"-k option: must specify a color name; '%s' is not a color name.\n",colorName);
			  fprintf(stderr,"%d color names are defined; look in %s\n",NumColors,profilename);
			  badopts++;

	gotcolor:	  break;

/* -l: location.  "-l 60/-120" centers map at Lat 60, Long -120  */
		case 'l': initcheck(initdone,c,&badopts);
			  nfields=sscanf(optarg,"%[^/]/%[^ \t]",arg1,arg2);
			  if (nfields<2) {
			      fprintf(stderr,"-l option: must specify both lat and long as -l 45.30/100.20\n");
			      badopts++;
			      break;
			  }
			  LatCenter=scanlat(arg1);
			  LongCenter=scanlat(arg2);
			  break;
/* -m: Binary map files follow. "-m f2 f2 f3"			 */
		case 'm': for(;optind<argc && argv[optind][0]!='-';optind++) {
			     if (!initdone) init(argv);
			     initdone++;
			     getmap(argv[optind]);
			  }
			  break;
/* -M: Canned PostScript map file 				*/
		case 'M': initcheck(initdone,c,&badopts);
			     getPSmap(optarg,argv);
			     initdone++;
			  break;
/* -n: nodename files follow. "-n f1 f2 f3" 			 */
		case 'n': if (badopts) exit(2);
			  for(;optind<argc && argv[optind][0]!='-';optind++) {
			     if (!initdone) init(argv); initdone++;
			     gethosts(argv[optind],FALSE,FALSE);
			  }
			  break;
/* -N: nodelabel options. "-N circle", "-Nd" or "-Nc", , etc.    	 */
		case 'N': switch (optarg[0]) {
			      case 'c':case 'C':HostLabels=Circles; break;
			      case 'd':case 'D':HostLabels=Dots; break;
			      case 't':case 'T':HostLabels=TrafficCircles; break;
			      case 'l':case 'L':HostLabels=TextDots; break;
			      case 'm':case 'M':HostLabels=TextCircles; break;
			  }
			  break;

/* -O: don't include technical labels on the maps */
		case 'O': initcheck(initdone,c,&badopts);
			  OmitDetails++;
			  break;

/* -p: projection. "-p l50/40" for Lambert; "-p m" for Mercator. */
		case 'p': initcheck(initdone,c,&badopts);
			  switch (optarg[0]) {
			  case 'l': case 'L':
			      Projection = Lambert;
			      sscanf(&optarg[1],"%lf/%lf", &Lamb1,&Lamb2);
			      break;
			  case 'm': case 'M':
			      Projection = Mercator;
			      break;
			  case 'g': case 'G':
			      Projection = Gall;
			      break;
			  case 'c': case 'C':
			      Projection = CentralCyl;
			      break;
			  default: fprintf(stderr,"Unrecognized argument to '-m' option (use L or M).\n");
			      break;
			  }
			  break;
/* -P: profile file. Save the argument as a file name to replace netmap.pro */
		case 'P': initcheck(initdone,c,&badopts);
			  strcpy(profilename,optarg);
			  break;
/* -r: rotate the map (positive is clockwise) after projecting */
		case 'r': initcheck(initdone,c,&badopts);
			  sscanf(optarg,"%d",&RotDegrees);
			  break;
/* -R: resolution. Dots per inch that the printer can show */
		case 'R': initcheck(initdone,c,&badopts);
		 	  sscanf(optarg, "%d",&DeviceResolution);
			  break;
/* -s: sieve size. Real argument is the detail threshhold for showing or not
       showing flow data */
		case 's': sscanf(optarg,"%lf",&Cutoff);
			  if (Cutoff < 0.0 || Cutoff > 101) Cutoff = 0.0;
			  break;

/* -S: stroke size. Integer argument is shortest vector that will be drawn. */
		case 'S': sscanf(optarg,"%d",&StrokeLimit);
			  MinimumStrokeSize = StrokeLimit;
			  MinimumStrokeSize *= MinimumStrokeSize; /* square */
			  break;
/* -t: Title of this map. "-t 'The North American Backbone' " */
		case 't': MapTitle=strcpy((char *)malloc(1+strlen(optarg)),
				          optarg);
			  break;

/* -T: Subtitle of this map. "-T 'string of your choice' " */
		case 'T': MapSubtitle=strcpy((char *)malloc(1+strlen(optarg)),
				          optarg);
			  break;

/* -w: window.  "-w 20/50.30" gives window 20 deg high 50 deg 30 min wide */
		case 'w': initcheck(initdone,c,&badopts);
			  nfields=sscanf(optarg,"%[^/]/%[^ \t]",arg1,arg2);
			  if (nfields<2) {
			      fprintf(stderr,"-w option: must specify both lat and long as -l 45.30/100.20\n");
			      badopts++;
			      break;
			  }
			  LatHeight=scanlat(arg1);
			  LongWidth=scanlat(arg2);
			  if (DebugLevel > 0) fprintf(stderr,
 "Lat/Long = %s/%s, scanned as %f/%f\n",arg1,arg2,LatHeight,LongWidth);
			  break;
/* -W: line width. Specify width of pen stroke for files read in now */
		case 'W': switch(optarg[0]) {
			      case '0': LineWidth=1; break;
			      case '1': LineWidth=2; break;
			      case '2': LineWidth=3; break;
			      case '3': LineWidth=4; break;
			      case '.': tc = optarg; switch(*++tc) {
				  case '0': LineWidth= -1; break;
				  case '1': LineWidth= -2; break;
				  case '2': LineWidth= -3; break;
				  case '3': LineWidth= -4; break;
			          default: fprintf(stderr,"-W. option: argument %s unrecognized; ignored.\n",optarg);
			      }
			      break;
			      case '-': tc = optarg; switch(*++tc) {
				  case '0': LineWidth= -5; break;
				  case '1': LineWidth= -6; break;
				  case '2': LineWidth= -7; break;
				  case '3': LineWidth= -8; break;
			          default: fprintf(stderr,"-W. option: argument %s unrecognized; ignored.\n",optarg);
			      }
			      break;
			      default: fprintf(stderr,"-W option: argument %s unrecognized; ignored.\n",optarg);
			  }
			  break;
/* -y: host-and-path files follow. "-y f1 f2 f3" 			 */
		case 'y': if (badopts) exit(2);
			  for(;optind<argc && argv[optind][0]!='-';optind++) {
			     if (!initdone) init(argv); initdone++;
			     gethosts(argv[optind],TRUE,FALSE);
			  }
			  break;
/* -z: host-flow-and-path files follow. "-z f1 f2 f3" 			 */
		case 'z': if (badopts) exit(2);
			  for(;optind<argc && argv[optind][0]!='-';optind++) {
			     if (!initdone) init(argv); initdone++;
			     gethosts(argv[optind],TRUE,TRUE);
			  }
			  break;

/* -Z: nondirectional flag for flow-and-path files. "-Z".		*/
		case 'Z': BiDirectional++;
			  break;

		case '?': exit(1);
	    }
	    fflush(stdout);
	    if (badopts) exit (2);
	    
	for(;optind<argc; optind++) {
	    if (!initdone) init(argv);
	    initdone++;
	    getmap(argv[optind]);
	}
	if (!initdone) init(argv);
	if (CannedMap) exit(0);
/* all done. Now dump the accumulated paths and nodes */
	ProcessNodes();
  	if (RawPathCount > 0) {
	    printf("%%%%Paths\n");
	    printf("PHcolor\n");
	    dumppaths();
	}
	printf("%%%%Nodes\n");
	printf("DTcolor\n");
	DrawNodes();
	if (SomeCircles && NumHosts > 0) {
	    printf("%%%%Circles\n");
	    printf("CLcolor\n");
	    dumpcircles();
	}
	if (DebugLevel > 0 && !UsingCanned)
	    fprintf(stderr,"Map has %d vectors; page draws %d vectors.\n",
		CumRawStrokeCount,CumStrokeCount);

	printf("LGcolor\n");
	if (MapTitle != NULL) printf("(%s) T1\n",MapTitle);
	if (MapSubtitle != NULL) printf("(%s) T2\n",MapSubtitle);

	printf("\n%%%%Trailer\nshowpage\n");

/*  (nodes have names now)  if (NumHosts > 0) dumpHosts();   */
	if (ErrorMessages) leftovers();
	printf("grestore\n");
	printf("grestore\n");
	exit(0);
}
