/*
 * This is a sample program, not used in netmap, that demonstrates how
 * to read the CBD files (compressed CIA map data) for use with other
 * programs. Because the CBD files are binary, you must pay attention to
 * byte ordering if you are going to create them on a machine using 
 * different byte ordering from the one you will read them on.
 */

#include <stdio.h>
#include <math.h>
#include "cbdmap.h"

/* ----------------------------------------------------------------------
   Read and process one binary map file.
 */

#include <sys/file.h>

main (argc, argv)
    int             argc;
    char          **argv;
{
int             Ifile, segcount, idx, idy, segbufsize, olt, oln, j, k, jstroke, iseg;

struct segdict
{
    BIT32           segid;
    BIT32           maxlat, minlat, maxlong, minlong;
    BIT32           absaddr;
    BIT16           nbytes;
    BIT16           rank;
}              *sd, *sdbuf;

struct segment
{
    BIT32           orgx, orgy;
    BIT32           id;
    BIT16           nstrokes;
    BIT16           dummy;
}               sb;

BIT16          *segbuf;

BIT32           i32;
char            fname[128];
struct cbdhead  header;

    strcpy (fname, argv[1]);

    Ifile = open (fname, O_RDONLY, 0);
    if (Ifile < 0)
    {
	perror (fname);
	return;
    }
    fprintf (stderr, "Processing map file %s\n", fname);

/*
 * Read in the file header, check it for the correct magic number,
 * and learn the address of the segment dictionary
 */
    read (Ifile, &header, sizeof header);

    twiddle32 (&(header.magic));
    twiddle32 (&(header.dictaddr));
    twiddle32 (&(header.segcount));
    twiddle32 (&(header.segsize));
    twiddle32 (&(header.segmax));

    if (header.magic != CBD_MAGIC)
    {
	fprintf (stderr, "File %s has bad magic number %d != %d\n",
		 fname, header.magic, CBD_MAGIC);
	exit (2);
    }

/* allocate space for the segment buffer */
    segbufsize = 2 * header.segmax;
    segbuf = (BIT16 *) malloc (50 + segbufsize);

/* allocate space for the segment dictionary */
    sdbuf = (struct segdict *) malloc (100 + header.segsize);
    sd = sdbuf;
    sd++;

/* Go read in the segment dictionary (it's at the end of the file) */
    lseek (Ifile, header.dictaddr, L_SET);
    j = read (Ifile, sd, header.segsize);

    if (j < header.segsize)
    {
	fprintf (stderr, "File %s: segment dictionary expecting %d got %d\n",
		 fname, header.segsize, j);
	close (Ifile);
	free (sdbuf);
	free (segbuf);
	return;
    }

/*
 * Now look at each segment and decide if we're
 * going to keep it or not
 */
    segcount = header.segcount;

    sd = sdbuf;
    for (iseg = 1; iseg <= segcount; iseg++)
    {
	sd++;			/* does this really work? wow! */

	twiddle32 (&(sd->segid));
	twiddle32 (&(sd->maxlat));
	twiddle32 (&(sd->minlat));
	twiddle32 (&(sd->maxlong));
	twiddle32 (&(sd->minlong));
	twiddle32 (&(sd->absaddr));
	twiddle16 (&(sd->nbytes));
	twiddle16 (&(sd->rank));

	lseek (Ifile, sd->absaddr, L_SET);
	read (Ifile, &sb, sizeof sb);
	twiddle32 (&(sb.orgx));
	twiddle32 (&(sb.orgy));
	twiddle32 (&(sb.id));
	twiddle16 (&(sb.nstrokes));

	if (sd->nbytes > segbufsize)
	{
	    fprintf (stderr, "%s: segment %d needs %d bytes; buffer limit is %d.\n", fname, iseg, sd->nbytes, segbufsize);
	    close (Ifile);
	    free (sdbuf);
	    free (segbuf);
	    return;
	}
	read (Ifile, segbuf, sd->nbytes);
	k = 0;
	oln = sb.orgx;
	olt = sb.orgy;

	printf ("m %f %f\n", olt / 36000., oln / 36000.);

	for (jstroke = 1; jstroke <= sb.nstrokes; jstroke++)
	{
	    twiddle16 (&segbuf[k]);
	    if (segbuf[k] & SHORTFLAG)
	    {
/* Flag bit on: unpack a 16-bit field into dx and dy */
		i32 = segbuf[k++];
		if (i32 > 0)
		    i32 &= ~SHORTFLAG;
		idy = i32 & 0xFF;
		if (idy & 0x80)
		    idy |= ~0xFF;	/* extend sign */
		idx = i32 / 256;
	    }
	    else
	    {
/* Flag bit off: take dx and dy from 32-bit fields. */
		idx = segbuf[k++];
		if (idx < 0)
		    idx |= SHORTFLAG;
		twiddle16 (&segbuf[k]);
		idx = (idx << 16) | segbuf[k];
		k++;

		twiddle16 (&segbuf[k]);
		idy = segbuf[k];
		k++;
		if (idy < 0)
		    idy |= SHORTFLAG;
		twiddle16 (&segbuf[k]);
		idy = (idy << 16) | segbuf[k];
		k++;
	    }
	    olt = olt + idy;
	    oln = oln + idx;
	    printf ("d %f %f\n", olt / 36000., oln / 36000.);
	}
    }
    close (Ifile);
    free (sdbuf);
    free (segbuf);
}

/*
 * Change the two routines below to accomodate the byte ordering on your
 * machine.
 */

twiddle32 (i32)
    BIT32          *i32;
{
BIT32           temp;

    temp = 0;
    temp |= 0xFF000000 & ((0x000000FF & *i32) << 24);
    temp |= 0x00FF0000 & ((0x0000FF00 & *i32) << 8);
    temp |= 0x0000FF00 & ((0x00FF0000 & *i32) >> 8);
    temp |= 0x000000FF & ((0xFF000000 & *i32) >> 24);
    *i32 = temp;
}

twiddle16 (i16)
    BIT16          *i16;
{
BIT16           temp;

    temp = 0;
    temp |= 0xFF00 & ((0x00FF & *i16) << 8);
    temp |= 0x00FF & ((0xFF00 & *i16) >> 8);
    *i16 = temp;
}
