/* 
   This file contains the code and data structures that are used to prevent
   the host dots from bumping into each other. We keep a matrix of slot
   numbers, and use that matrix for host and arc placement guidance. When a
   host occupies a slot, the id number of the host is stored in that cell of
   the matrix.

	Brian K. Reid, DEC Western Research, September 1986
 */

#include "netmap.h"

short	*cmap;
double	Xslotinc,		/* X value of one slot */
	Yslotinc;		/* Y value of one slot */
int	XSlots,			/* How many X slots? */
	YSlots;			/* How many Y slots? */

/* return slot numbers corresponding to x and y positions */
int xslot(x) double x; {
    return(1+(int) 0.5+(XSlots*((x-ClipXmin)/(ClipXmax-ClipXmin))));
}

int yslot(y) double y; {
    return(1+(int) 0.5+(YSlots*((y-ClipYmin)/(ClipYmax-ClipYmin))));
}

/* return x and y positions corresponding to rounded integer slot numbers */
double slotx(x) double x; {
    int j; double jx;
    j = xslot(x)-1; jx = (Xslotinc*j) + ClipXmin;
    return(jx+(Xslotinc/2.0));
}

double sloty(y) double y; {
    int j; double jy;
    j = yslot(y)-1; jy = (Yslotinc*j) + ClipYmin;
    return(jy+(Yslotinc/2.0));
}



int xspiral [] = {1,-1,0, 0, 1, 1,-1,-1, 2, -2, 0, 0, 2,-2, 2,-2, 2,-2, 2,-2};
int yspiral [] = {0, 0,1,-1,-1, 1,-1, 1, 0, 0, -2, 2, 1, 1,-1,-1,-2, 2, 2, 2};
#define Nspiral	20

/* ----------------------------------------------------------------------- */

/* This function finds an [x,y] slot that is not currently occupied and moves
   [Hx,Hy] to that slot. If it fails to find a slot, it leaves Hx and Hy
   unchanged. 
   
   If constrain=1, only x-direction motion is tried; if contrain = -1,
   only y-direction motion is tried.
 */

SlotLock(HostName,Hx,Hy,constrain)
char *HostName;
double *Hx,*Hy;
int constrain;
{
    int j,sn;
    if (DebugLevel > 3) fprintf(stderr,"\n%s: looking for slot.\n",HostName);
    if (slot(*Hx,*Hy)) {

	if (constrain) {
	    for(j=2; j<20; j += 2) {
		for (sn= -1; sn <= 1; sn += 2) {
		    if (constrain == 1) {
			if (*Hx + Xslotinc*j*sn > ClipXmax+Xslotinc) break;
			if (*Hx + Xslotinc*j*sn < ClipXmin-Xslotinc) break;
			if (!slot(*Hx + Xslotinc*j*sn,*Hy)) {
			    *Hx = *Hx + Xslotinc*j*sn;
			    goto done;
			}
		    } else {
		        if (*Hy + Yslotinc*j*sn > ClipYmax+Yslotinc) break;
		        if (*Hy + Yslotinc*j*sn < ClipYmin-Yslotinc) break;
			if (!slot(*Hx, *Hy + Yslotinc*j*sn)) {
			    *Hy = *Hy + Yslotinc*j*sn;
			    goto done;
			}
		    }
		}
	    }
	} else {
	    for(j = 0; j < Nspiral; j++) {
		if (!slot(*Hx + Xslotinc*xspiral[j],
			  *Hy + Yslotinc*yspiral[j])) {
		    *Hx = *Hx + Xslotinc*xspiral[j];
		    *Hy = *Hy + Yslotinc*yspiral[j];
		    goto done;
		}
		if (DebugLevel > 4) fprintf(stderr,"%s: tried [%d,%d]\n",HostName,
		    	xslot(*Hx + Xslotinc*xspiral[j]),
		    	      yslot(*Hy + Yslotinc*yspiral[j]));
	    }
	    for(j = 0; j < Nspiral; j++) {
		if (!slot(*Hx + Xslotinc*2*xspiral[j],
			  *Hy + Yslotinc*2*yspiral[j])) {
		    *Hx = *Hx + Xslotinc*2*xspiral[j];
		    *Hy = *Hy + Yslotinc*2*yspiral[j];
		    goto done;
		}
	    }
	}
    if (DebugLevel > 3) fprintf(stderr,"No slot for %s([%f,%f] %d %f %f\n",
    		   HostName,*Hx,*Hy,constrain,Xslotinc,Yslotinc);
    }
done: ;
}
