/* 
   This file contains various functions that are called at the end of
   execution to wrap up one thing or another.
 */

#include "netmap.h"

    double Zhostmax;
    static int hcompare();
/* ----------------------------------------------------------------------- */

/* Postprocess the node table. Depending on options, we want to draw the
   nodes as dots or as circles. 
*/

ProcessNodes()
{
    int i,imax;
    
    double tpercent,Hx,Hy;

    if (DebugLevel>2) fprintf(stderr,"Processing %d nodes\n",NumHosts);
/* First find the largest host traffic value, so we can do arithmetic in
   percentiles. */
    Zhostmax = 0.0; imax = 0;
    for(i=1; i<=NumHosts; i++) {
	tpercent = htable[i].inTraf;
	if (Zhostmax < tpercent) {Zhostmax = tpercent; imax = i;}
	if (DebugLevel > 6 && htable[i].inbounds != 0) {
	    fprintf(stderr,"%9s  %9.2f\n", htable[i].hname, tpercent);
	}
    }
    if (DebugLevel > 0) fprintf(stderr,"Maximum host traffic was %3.1f at %s\n",Zhostmax,htable[imax].hname);

/* If there are any nodes whose "drawn" field flags them as
   "DecideByTraffic", then make that decision now, and convert them either to
   dots or to circles. */

    if (Zmax) for(i=1; i<=NumHosts; i++) {
	if (htable[i].inbounds != 0 &&
	   htable[i].drawn == DecideByTraffic) {
	    if (htable[i].inTraf > 0 || htable[i].outTraf > 0) {
		tpercent = (100 * htable[i].inTraf) / Zhostmax;
		if (tpercent >= Cutoff) {
		    htable[i].drawn = DrawCircle;
		} else {
		    htable[i].drawn = DrawDot;
		}
	    } else {
	        htable[i].drawn = NodeDrawn; /* fib so it won't get drawn */
	    }
	}
    }

/* Now force every circle node into a grid slot of one kind or another, so
   that they won't all land on top of each other. */
    for(i=1; i<=NumHosts; i++) {
	if (htable[i].drawn == DrawCircle
	 || htable[i].drawn == DrawTextCircle) {
	    Hx = htable[i].hstx;
	    Hy = htable[i].hsty;
	    SlotLock(htable[i].hname,&Hx,&Hy,0);
	    slot(Hx,Hy)=i;
	    htable[i].hstx = slotx(Hx);
	    htable[i].hsty = sloty(Hy);
	}
    }
}
/* ----------------------------------------------------------------------- */

/* draw the nodes that are not going to rate full-blown circles. First
   sort them into order by X and Y position so that duplicates will be
   adjacent, so we can press out duplicates and use delta positioning
   on the ones that aren't duplicates. 
*/
DrawNodes()
{

    int j,i,textx,texty,textangle,FinitDone;
    int wasX,wasY,deltaX,deltaY,firstDot,rowcount;
    int (*htptr)[1];

    FinitDone = FALSE;
    htptr = (int (*)[1]) malloc(sizeof (int) * (1+NumHosts));
    for (j=0; j < NumHosts; j++) {
	(*htptr)[j] = j+1;
    }
    qsort(htptr,NumHosts,sizeof(int),hcompare);
    firstDot = TRUE;
    rowcount = 0;
    for(j=0; j < NumHosts; j++) {
        i = (*htptr)[j];
        if (htable[i].inbounds != 0) {
 	    if (htable[i].drawn == DrawDot) {
		if (firstDot) {
		    PSP(htable[i].hstx,htable[i].hsty,"Hd");
		    wasX = htable[i].hstx;
		    wasY = htable[i].hsty;
		    firstDot = FALSE;
		} else { 
		    deltaX = htable[i].hstx - wasX;
		    deltaY = htable[i].hsty - wasY;
		    if ((deltaY != 0) || (deltaX != 0)) {
		        if (deltaX == 0) {
			    printf("%d Hx",deltaY);
			} else {
			    printf("%d %d Hr",deltaX,deltaY);
			}
			if (rowcount++ < 6) printf(" ");
			    else {rowcount = 0; printf("\n"); }
			wasX = wasX + deltaX;
			wasY = wasY + deltaY;
		    }
		}

	        htable[i].drawn = NodeDrawn;
	    } else if (htable[i].drawn == DrawTextDot) {
	        if (!FinitDone) {
		    printf("/HFn HF def\n");
		    FinitDone = TRUE;
		}
		textx = 7;
		PlanCaption(i,&textx,&texty,&textangle);
		printf("(%s) %d %d %d ",
		    htable[i].hname,textx,texty,textangle);
	        PSP(htable[i].hstx,htable[i].hsty,"Id");
	        htable[i].drawn = NodeDrawn;
	    }
	}
    }
    if (rowcount != 0) printf("\n");

/* Note that we wait to draw the circles until after the paths have been
   drawn, so that the ends of the paths don't make the circles ugly. */
}

static int hcompare (rec1,rec2)
int *rec1,*rec2;
{
   int ix,iy,jx,jy;
   ix = htable[*rec1].hstx; iy = htable[*rec1].hsty;
   jx = htable[*rec2].hstx; jy = htable[*rec2].hsty;
   if (ix<jx) return (-1);
   if (ix>jx) return (1);
   if (iy<jy) return (-1);
   if (iy>jy) return(1);
   return(0);
}

/* ----------------------------------------------------------------------- */

/* See if there were any significant relay hosts left out of the diagram. */
leftovers()
{
    double tpercent,x,y,Ymin,Ymax,Xmin,dx,dy;
    int i,j,virgin,NumListed;
    char Flag;

    struct hn *(*s)[500];	/* 500 is to keep dbx and saber happy */
    struct hn *t;

    Ymin = 30; Ymax = ImageHeight-10;  dy = 10; dx = 140;
    Xmin = 15; 
    virgin = TRUE;

    if (Zmax == 0) return;
    NumListed = 0;
    for(i=1; i<=NumHosts; i++) {
	tpercent = (100 * htable[i].outTraf) / Zhostmax;
        if (tpercent >= Cutoff) NumListed++;
    }
/* Now make an index array that points to the above-threshhold nodes */
    s = (struct hn *(*)[500]) malloc(sizeof(t) * (1+NumListed));
    j = 0;
    for (i=1; i<=NumHosts; i++) {
	tpercent = (100 * htable[i].outTraf) / Zhostmax;
        if (tpercent >= Cutoff) {
	   j++;
	   (*s)[j] = &(htable[i]);
	}
    }

/* Now sort the index into descending order by traffic percent */
    for (i=1; i<NumListed; i++) {
       for (j=i+1; j<=NumListed; j++) {
           if ( ((*s)[j])->outTraf > ((*s)[i])->outTraf ) {
	       t = (*s)[j]; (*s)[j] = (*s)[i]; (*s)[i]=t;
	   }
       }
    }

    for(i=1; i<=NumListed; i++) {
	tpercent = (100 * ((*s)[i])->outTraf) / Zhostmax;
	if (virgin) {
	    y = Ymax + dy; x = Xmin;
	    printf("\nINIT\n");
	    if (MapTitle != NULL) printf("(%s) T1\n",MapTitle);
	    printf("(Percent of your node's traffic carried by other nodes. * indicates a node whose location is unknown.) T2\n");
	    printf("(DECWRL %s at %s) 1 MLB\n",ProgID,sdat);
	    printf("HostNameFont CaptionSize CF \n");
	    virgin = FALSE;
	}
	y -= dy;
	if (y < Ymin) {
	    y = Ymax;
	    x += dx;
	}
	PSP(x,y,"MT ");
	Flag = ' '; if (! ((*s)[i])->datacomplete) Flag = '*';
	printf("(%c%-10s  %6.2f%%) show\n",Flag,((*s)[i])->hname,tpercent);
	if (DebugLevel > 0)
	    fprintf(stderr,"Host %s carries %3.1f%% of the traffic\n",((*s)[i])->hname,tpercent);
    }
    if (virgin) return;

    printf("\nshowpage\n");
}

/* ----------------------------------------------------------------------- */

PlanCaption(inode,ndx,ndy,ntheta)
/*ARGSUSED*/
   int inode,*ndx,*ndy,*ntheta;
{
    *ndy = -4;
    *ntheta = CaptionAngle+45;
}

/* ----------------------------------------------------------------------- */

/* This data structure is used to alphabetically index the host table */

    static struct hn *(*s)[500];	/* 500 is to keep dbx happy */
    static ncircles;
/* ----------------------------------------------------------------------- */


/* If nodes are labeled with numbered circles, draw them after the paths go
   down so that the paths do not obliterate the circle numbers. */

dumpcircles () {
    int idno,i,j,
	textx,texty,textangle;
    struct hn *t;

    ncircles=0;

/* First count the circles and mark them as counted. We don't use the
   "inbounds" flag because some of the hosts may have been moved in bounds by
   "movehost", but that flag will be left false so that paths don't get drawn
   between two out-of-bounds hosts.  
 */
    for(i=1; i<=NumHosts; i++) {
	if (clipxy(htable[i].hstx,htable[i].hsty)) {
	    if (htable[i].drawn ==DrawCircle
	     || htable[i].drawn==DrawTextCircle)
		ncircles++;
	}
    }

    if (DebugLevel > 1) fprintf(stderr,"Found %d circles.\n",ncircles);
    if (ncircles == 0) return;

/* Now make an index array that points to the allocated circles */
    s = (struct hn *(*)[500]) malloc(sizeof(t) * (1+ncircles));
    j = 0;
    for (i=1; i<=NumHosts; i++) {
        if ((htable[i].drawn ==DrawCircle
	 || htable[i].drawn==DrawTextCircle)
	 && clipxy(htable[i].hstx,htable[i].hsty)) {
	   j++;
	   (*s)[j] = &(htable[i]);
	}
    }

/* Now sort the index into alphabetical order by node name */
    for (i=1; i<ncircles; i++) {
       for (j=i+1; j<=ncircles; j++) {
           if (strcmp( ((*s)[j])->hname, ((*s)[i])->hname) < 0) {
	       t = (*s)[j]; (*s)[j] = (*s)[i]; (*s)[i]=t;
	   }
       }
    }

/* Now assign numbers in alphabetical order and print them 
   6 May 1987: changed to print names instead of numbers. Left the
   number code in because it was tricky to write and doesn't cost
   much cpu time.
 */
    printf("HF setfont gsave /HFn IF def\n");
    idno = 0;     

    for (j=1; j<=ncircles; j++) {
        if (((*s)[j])->drawn ==DrawCircle) {
   	    ((*s)[j])->hid = ++idno;
	    printf("(%s) %1.0f ",((*s)[j])->hname,CaptionAngle);
	    PSP(((*s)[j])->hstx,((*s)[j])->hsty,"HST");
	}
        if (((*s)[j])->drawn ==DrawTextCircle) {
   	    ((*s)[j])->hid = ++idno;
	    textx = 14;
	    PlanCaption((*s)[j]->hid,&textx,&texty,&textangle);
	    printf("(%s) %d %d %d (%d) %1.0f ",(*s)[j]->hname,textx,
	    	texty,textangle,((*s)[j])->hid,CaptionAngle);
	    PSP(((*s)[j])->hstx,((*s)[j])->hsty,"HTT");
	}
    }
    printf("grestore\n");
}

/* ----------------------------------------------------------------------- */

/* Print out a separate page with the list of host names and numbers */

dumpHosts() {
    double x,y,Ymin,Ymax,Xmin,Xmax,dx,dy;
    int i,virgin;

    if (!SomeCircles) return;
    printf("%%%%Nodeindex\n");
    Ymin = 30; Ymax = ImageHeight-10;  dy = 10; dx = 80;
    Xmin = 15; Xmax = ImageWidth-dx; 
    virgin = TRUE;

    y = Ymin-1; x = Xmax + 1;
    for (i=1; i<=ncircles; i++) {
	y -= dy;
	if (y < Ymin) {
	    y = Ymax;
	    x += dx;
	    PSP(x,y,"HSHmt");
	    if (x > Xmax) {
	        if (!virgin) printf("showpage\n");
		printf("\nINIT\n");
		if (MapTitle != NULL) printf("(%s) T1\n",MapTitle);
		printf("(Index to node numbers on map page) T2\n");
		printf("(DECWRL %s at %s) 1 MLB\n",ProgID,sdat);
		printf("HostNameFont CaptionSize CF \n");
		x = Xmin;
		PSP(x,y,"HSHmt");
		PSP(dx,dy,"HSHdxy");
	    }
	}
	printf("(%4d  %-10s) HSH\n",
		(*s)[i]->hid,(*s)[i]->hname);
	virgin = FALSE;
    }
    printf(" showpage\n");
}

